/**
 * ShimmerSimService.java
 * Copyright (C) 2012 Pattern Recognition Lab, University Erlangen-Nuremberg.
 *  
 * Licensed under the Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License (the "License");
 * you may not use this file except in compliance with the License.
 * To view a copy of this License, visit http://creativecommons.org/licenses/by-nc-sa/3.0/.
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on 
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the 
 * specific language governing permissions and limitations under the License.
 * 
 * 
 * This file is part of the "Hearty" Android Application. It was released as supplementary material related to the publication [1]:
 * [1] S. Gradl, P. Kugler, C. Lohmüller, and B. Eskofier, “Real-time ECG monitoring and arrhythmia detection using Android-based mobile devices,” in 34th Annual International Conference of the IEEE EMBS, 2012, pp. 2452–2455.
 * 
 * It is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * If you reuse this code you have to keep or cite this comment.
 *
 */
package de.lme.hearty;


import java.io.File;
import java.util.ArrayList;

import android.app.Service;
import android.content.Intent;
import android.os.DeadObjectException;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.IBinder;
import android.os.Looper;
import android.os.Message;
import android.os.RemoteException;
import android.util.Log;
import de.lme.hearty.ShimmerProcessor.OnShimmerProcessorEventListener;
import de.lme.plotview.DataSource;
import de.lme.plotview.Plot;
import de.lme.plotview.Plot.PlotStyle;
import de.lme.plotview.SamplingPlot;




/**
 * @author Stefan Gradl; Patrick Kugler
 */
public class ShimmerSimService extends Service implements OnShimmerProcessorEventListener
{
	public static final String					INTENT_EXTRA_SHIM_MODE		= "de.lme.shimmode";
	public static final String					INTENT_EXTRA_DATA_SOURCE	= "de.lme.datasource";
	public static final String					INTENT_EXTRA_DATA_COL		= "de.lme.col";
	public static final String					INTENT_EXTRA_DATA_MULT		= "de.lme.mult";
	public static final String					INTENT_EXTRA_DATA_COUNT		= "de.lme.simcount";

	public static final int						SHIM_MESSAGE_INIT			= 0;
	public static final int						SHIM_MESSAGE_SHUTDOWN		= 1;
	public static final int						SHIM_MESSAGE_EVENT			= 2;
	public static final int						SHIM_MESSAGE_SIM_EVENT		= 3;
	public static final int						SHIM_MESSAGE_SIM_END		= 4;

	public static final int						SHIM_MODE_LIVE				= 0;
	public static final int						SHIM_MODE_SIMULATION		= 1;
	public static final int						SHIM_MODE_INTERN			= 2;
	public static final int						SHIM_MODE_BLUETOOTH			= 3;
	public static final int						SHIM_MODE_INVALID			= 4;

	private static IShimmerSimServiceCallback	m_callback;


	private SamplingPlot						simPlot						= null;
	private int									simPlotIter					= 0;
	/** Post delay in milliseconds */
	private float								simDelay					= 5f;
	/** Virtual simulation time */
	private float								simVirtualTime				= 0f;
	/** Regular simulation time */
	private long								simTime						= 0;


	public static class ShimConnection
	{
		/** Simulation mode */
		public int		simMode			= SHIM_MODE_BLUETOOTH;

		/** Record String ob MIT-BIH record file */
		public String	dataSource		= "default";
		/** Data column in record file */
		public int		dataColumn		= 0;
		/** Multiplier */
		public int		dataMultiplier	= 1;
		/** Number of samples to simulate */
		public int		simCount		= 0;
	}


	public static ShimConnection			m_con		= new ShimConnection();

	private ArrayList< ShimmerDataSource >	mShimmerDataSources;

	private DisplayProcessor				m_displayProcessor;


	/** WFDB signal samples */
	public static DataSource.wfdbEcgSignal	simSignal	= null;


	/* (non-Javadoc)
	 * @see android.app.Service#onBind(android.content.Intent)
	 */
	@Override
	public IBinder onBind (Intent intent)
	{
		m_con.simMode = intent.getIntExtra( INTENT_EXTRA_SHIM_MODE, SHIM_MODE_BLUETOOTH );
		if (m_con.simMode >= SHIM_MODE_INVALID || m_con.simMode < 0)
		{
			return null;
		}

		m_con.dataSource = intent.getStringExtra( INTENT_EXTRA_DATA_SOURCE );
		if (m_con.dataSource == null)
		{
			return null;
		}

		m_con.dataColumn = intent.getIntExtra( INTENT_EXTRA_DATA_COL, 2 );
		m_con.dataMultiplier = intent.getIntExtra( INTENT_EXTRA_DATA_MULT, 1000 );
		m_con.simCount = intent.getIntExtra( INTENT_EXTRA_DATA_COUNT, 0 );

		return mBinder;
	}


	/* (non-Javadoc)
	 * @see android.app.Service#onUnbind(android.content.Intent)
	 */
	@Override
	public boolean onUnbind (Intent intent)
	{
		// === TODO Auto-generated method stub
		return super.onUnbind( intent );
	}


	private Looper			mServiceLooper;
	private ServiceHandler	mServiceHandler;
	private Thread			m_loadingThread	= null;
	private HandlerThread	m_serviceThread	= null;


	/* (non-Javadoc)
	 * @see android.app.Service#onCreate()
	 */
	@Override
	public void onCreate ()
	{
		// === TODO Auto-generated method stub
		super.onCreate();

		if (m_con == null)
			return;

		m_loadingThread = new Thread( new Runnable() {
			public void run ()
			{
				// ==============================================
				// == SIMULATED INTERNAL TEST DATA
				// ====>
				if (m_con.simMode == SHIM_MODE_INTERN)
				{
					// load internal test record
					simPlot = new SamplingPlot( "ECG", Plot.generatePlotPaint(), PlotStyle.LINE, 20000 );
					// simPlot.loadFromFile( this, this.getResources().openRawResource( R.raw.ecg_ex ) );
					simPlot.loadFromFile(	ShimmerSimService.this,
											ShimmerSimService.this.getResources().openRawResource( R.raw.ecgpvc ) );
					simDelay = 5f;

					m_con.simCount = simPlot.values.num;
				}
				// <====
				// ==============================================

				// ==============================================
				// == SIMULATED MIT-BIH DATA
				// ====>
				else if (m_con.simMode == SHIM_MODE_LIVE || m_con.simMode == SHIM_MODE_SIMULATION)
				{
					File sigFile = new File( ShimmerSimService.this.getExternalFilesDir( null ), "mitdb" + m_con.dataSource
							+ "sig.csv.gz" );
					File annFile = new File( ShimmerSimService.this.getExternalFilesDir( null ), "mitdb" + m_con.dataSource
							+ "ann.csv" );

					// load MIT-BIH Database record if not already in memory
					// if (simSignal == null || !simSignal.recordName.equals( sigFile.getAbsolutePath() ))
					{
						// m_simCount = 10000;
						Log.d( "ShimService", "Loading... " + m_con.simCount + " " + m_con.dataColumn + " "
								+ m_con.dataMultiplier + " " + sigFile.getAbsolutePath() );
						// load data
						simSignal = new DataSource.wfdbEcgSignal();
						if (simSignal.load( sigFile.getAbsolutePath(),
											annFile.getAbsolutePath(),
											m_con.simCount,
											m_con.dataColumn,
											m_con.dataMultiplier ) != 0)
						{
							Log.d( "ShimService", "error loading file!" );
							return;
						}

						if (m_callback == null || simSignal == null)
						{
							ShimmerSimService.this.stopSelf();
							return;
						}

						// calculate sampling delay
						simDelay = simSignal.sampleInterval * 1000;

						if (m_con.simCount <= 0 || m_con.simCount > simSignal.ml2.num)
							m_con.simCount = simSignal.ml2.num;
					}

				}
				// <====
				// ==============================================

				// ==============================================
				// == BLUETOOTH
				// ====>
				else if (m_con.simMode == SHIM_MODE_BLUETOOTH)
				{

					mShimmerDataSources = new ArrayList< ShimmerDataSource >();
					{
						// String[] shimmerIDs = {"FireFly-E050"};
						String[] shimmerIDs = { "RN42-1EE4" };
						ShimmerManager shimmerManager = new ShimmerManager();
						// for (String shimmerID : shimmerIDs)
						// {
						// // Log at least an error if the device to connect to is not
						// // available
						// if (shimmerManager.getShimmerDeviceByID( shimmerID ) == null)
						// {
						// Log.e( "ShimmerDisplay", "Cannot find device " + shimmerID + " in list of paired devices." );
						// }
						// mShimmerDataSources.add( shimmerManager.getShimmerDeviceByID( shimmerID ) );
						// }

						mShimmerDataSources.add( shimmerManager.getShimmerDeviceByID( m_con.dataSource ) );

						ShimmerSimService.this.m_displayProcessor = new DisplayProcessor( mShimmerDataSources );
						m_displayProcessor.setOnShimmerProcessorEventListener( ShimmerSimService.this );


					}

				}
				// <====
				// ==============================================


				try
				{
					// inform about sampling rate
					if (m_callback == null)
					{
						ShimmerSimService.this.stopSelf();
						return;
					}

					m_callback.onShimmerSimEvent( SHIM_MESSAGE_INIT, (long) (1000d / simDelay), 0, (char) 0 );
				}
				catch (RemoteException e)
				{
					if (DeadObjectException.class.isInstance( e ))
					{
						// remove the dead callback
						m_callback = null;
						ShimmerSimService.this.stopSelf();
						// don't start
						return;
					}
				}


				simPlotIter = 0;
				simTime = 0;
				simVirtualTime = 0f;


				m_serviceThread = new HandlerThread( "ShimmerSimService", android.os.Process.THREAD_PRIORITY_MORE_FAVORABLE );
				m_serviceThread.start();

				// Get the HandlerThread's Looper and use it for our Handler
				mServiceLooper = m_serviceThread.getLooper();
				mServiceHandler = new ServiceHandler( mServiceLooper );

				// send initial message
				mServiceHandler.sendEmptyMessage( REPORT_MSG );
			}
		} );
		m_loadingThread.start();


		// Thread.currentThread().setPriority( Thread.MAX_PRIORITY );
		// mHandler.sendEmptyMessage( REPORT_MSG );
	}


	// Handler that receives messages from the thread
	private final class ServiceHandler extends Handler
	{
		public ServiceHandler (Looper looper)
		{
			super( looper );
		}


		@Override
		public void handleMessage (Message msg)
		{
			switch (msg.what)
			{
				case REPORT_MSG:

					try
					{
						// timer = System.nanoTime();

						if (m_callback == null)
							return;

						simTime += simDelay;


						// Send regular simulation events
						if (m_con.simMode == SHIM_MODE_LIVE)
						{
							if (m_callback != null)
								m_callback.onShimmerSimEvent(	SHIM_MESSAGE_SIM_EVENT,
																simTime,
																simSignal.ml2.values[ simPlotIter ],
																(char) simSignal.labels.get( simPlotIter, 0 ) );
						}
						else if (m_con.simMode == SHIM_MODE_INTERN)
						{
							if (m_callback != null)
								m_callback.onShimmerSimEvent(	SHIM_MESSAGE_SIM_EVENT,
																simTime,
																simPlot.values.values[ simPlotIter ],
																(char) 0 );
						}

						// Simulate entire record
						else if (m_con.simMode == SHIM_MODE_SIMULATION)
						{
							// non-real-time
							for (simPlotIter = 0; simPlotIter < m_con.simCount; ++simPlotIter)
							{
								if (m_callback == null)
									return;

								simVirtualTime += simDelay;
								simTime = (long) simVirtualTime;

								m_callback.onShimmerSimEvent(	SHIM_MESSAGE_SIM_EVENT,
																simTime,
																simSignal.ml2.values[ simPlotIter ],
																(char) simSignal.labels.get( simPlotIter, 0 ) );

							}
							m_callback.onShimmerSimEvent( SHIM_MESSAGE_SIM_END, 0, 0, (char) 0 );
							return;
						}
					}
					catch (RemoteException e)
					{
						if (DeadObjectException.class.isInstance( e ))
						{
							// remove the dead callback
							m_callback = null;
							if (mServiceLooper != null)
								mServiceLooper.quit();
							m_con = new ShimConnection();
							simSignal = null;
							ShimmerSimService.this.stopSelf();
						}
					}
					catch (NullPointerException e)
					{
						e.printStackTrace();
						m_callback = null;
						if (mServiceLooper != null)
							mServiceLooper.quit();
						m_con = new ShimConnection();
						simSignal = null;
						ShimmerSimService.this.stopSelf();
					}


					++simPlotIter;


					switch (m_con.simMode)
					{
						case SHIM_MODE_INTERN:
						case SHIM_MODE_LIVE:
							if (simPlotIter < m_con.simCount)
							{
								sendMessageDelayed( obtainMessage( REPORT_MSG ), (long) simDelay );
							}
							break;

						default:
							break;
					}

					break;

				default:
					super.handleMessage( msg );
			}
		}
	}


	private final IShimmerSimService.Stub	mBinder		= new IShimmerSimService.Stub() {

															public void registerCallback (IShimmerSimServiceCallback cb)
																	throws RemoteException
															{
																m_callback = cb;
															}


															public void unregisterCallback (IShimmerSimServiceCallback cb)
																	throws RemoteException
															{
																m_callback = null;
																if (mServiceLooper != null)
																	mServiceLooper.quit();
																simPlotIter = 0;
																m_con = new ShimConnection();
																simSignal = null;
																ShimmerSimService.this.stopSelf();
															}
														};

	private static final int				REPORT_MSG	= 1;


	/* (non-Javadoc)
	 * @see de.lme.hearty.ShimmerProcessor.OnShimmerProcessorEventListener#onShimmerProcessorEvent()
	 */
	public void onShimmerProcessorEvent ()
	{
		// Log.d( "d", "data" );

		try
		{
			m_callback.onShimmerSimEvent( SHIM_MESSAGE_SIM_EVENT, 0, // timestamp
											0, // value
											(char) 0 // label (acc. to MIT-BIH)
					);
		}
		catch (RemoteException e)
		{
			if (DeadObjectException.class.isInstance( e ))
			{
				// remove the dead callback
				m_callback = null;
				if (mServiceLooper != null)
					mServiceLooper.quit();
			}
		}
		catch (NullPointerException e)
		{
			e.printStackTrace();
			m_callback = null;
			if (mServiceLooper != null)
				mServiceLooper.quit();
		}

	}

}
